package main;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;

public class RecordedGIOutputProcessor {
	public static void loadData() {
		ArrayList<ArrayList<String>> allData = new ArrayList<ArrayList<String>>();

		File folder = new File("./RawGIData");
		
		for (File file:folder.listFiles()) {
			if (file.isFile() && file.getName().contains(".log") && file.getName().contains("FitnessData")) {
				//		    	System.out.println("Found: " + file);

				try {
					ArrayList<String> local = new ArrayList<String>();
					allData.add(local);
					FileReader fileReader = new FileReader(file);
					BufferedReader bufferedReader = new BufferedReader(fileReader);
					String line;
					while ((line = bufferedReader.readLine()) != null) {
						local.add(line);
					}
					fileReader.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}

		ArrayList<ArrayList<RecordedGeneration>> asGenArrays = new ArrayList<ArrayList<RecordedGeneration>>();
		for (ArrayList<String> file:allData) {
			ArrayList<RecordedGeneration> episode = new ArrayList<RecordedGeneration>();
			asGenArrays.add(episode);

			double bestEver = 0;
			for (String line : file) {
				RecordedGeneration currentGen = new RecordedGeneration();
				episode.add(currentGen);

				ArrayList<String> items = new ArrayList<String>();
				String working = "";
				boolean writing = false;
				for (char c:line.toCharArray()) {
					if (c == '[') {
						writing = true;
						working = "" + c;
					}
					else if (c == ']') {
						writing = false;
						working += c;
						items.add(working);
						currentGen.pop.add(new RecordedIndividual(working));
						working = null;
					}
					else if (writing) {
						working += c;
					}
				}
				if (currentGen.bestCorrectness() > bestEver) {
					bestEver = currentGen.bestCorrectness();
				}
				currentGen.bestEver = bestEver;
			}
		}
	}
	
	public void outputToConsole(ArrayList<ArrayList<RecordedGeneration>> asGenArrays) {
		printAvrgBest(asGenArrays);
	}

	/**
	 * Returns every single individual, from every single generation.
	 * Used for in-depth analysis of genetic improvement across entire populations, rather than
	 * simply the selected best
	 */
	public static void printAllFromGeneration(ArrayList<ArrayList<RecordedGeneration>> asGenArrays) {
		int minSize = asGenArrays.get(0).size();
		for (ArrayList<RecordedGeneration> episode:asGenArrays) {
			if (episode.size() < minSize) {
				minSize = episode.size();
			}
		}


		for (int i=0;i<minSize;i++) {
			System.out.print("Gen: " + i + " ");
			for (ArrayList<RecordedGeneration> episode:asGenArrays) {
				RecordedGeneration generation = episode.get(i);
				for (RecordedIndividual indiv:generation.pop) {
					int iVal = (int)(100000*indiv.correctness);
					double dVal = iVal;
					dVal /= 1000;
					System.out.print(" " + dVal + "%");
				}
			}
			System.out.println();
		}
	}
	
	/**
	 * Prints the average "Best fitness" for a number of files.
	 * This method is designed to allow easy assessment of the code's performance, as it
	 * compresses multiple runs' results into a single fitness signal
	 */
	public static void printAvrg(ArrayList<ArrayList<RecordedGeneration>> asGenArrays) {
		int minSize = asGenArrays.get(0).size();
		for (ArrayList<RecordedGeneration> episode:asGenArrays) {
			if (episode.size() < minSize) {
				minSize = episode.size();
			}
		}


		for (int i=0;i<minSize;i++) {
			System.out.print("Gen: " + i + " ");
			for (ArrayList<RecordedGeneration> episode:asGenArrays) {
				int iVal = (int)(100000*episode.get(i).bestEver);
				double dVal = iVal;
				dVal /= 1000;
				System.out.print(" " + dVal + "%");
			}
			System.out.println();
		}
	}
	
	/**
	 * Returns the average fitness for each individual within each generation
	 * Used to assess how fitness improved population-wide for multiple experiment runs
	 */
	public static void printAvrgAvrg(ArrayList<ArrayList<RecordedGeneration>> asGenArrays) {

		double[] sumAvrg = new double[100];
		int nUsed = 0;
		for (ArrayList<RecordedGeneration> episode:asGenArrays) {
			if (episode.size() < 100) {continue;}
			nUsed += 1;
			for (int i=0;i<100;i++) {
				sumAvrg[i] += episode.get(i).avrgCorrectness();
			}
		}

		for (int i=0;i<sumAvrg.length;i++) {
			sumAvrg[i] /= nUsed;
			System.out.println("Gen: " + i + "," + sumAvrg[i]);
		}
	}

	/**
	 *
	 * As with printAvrg, this compressess multiple files down into a single output, of their
	 * best fitness achieved, to allow plotting of the code's overall average progress.
	 * 
	 * This version allows "overrunning", it prints generation counts up to the highest count seen.
	 * This means that if some runs are shorter than others they will be excluded from the average
	 * at and after the point where they have no data.
	 * 
	 * This means that the later values might not be formed from all run data, but only a subset thereof
	 */
	public static void printAvrgBestOverrun(ArrayList<ArrayList<RecordedGeneration>> data) {
		int gen = 0;
		outer:
			while(true) {
				double sum = 0;
				int count = 0;
				boolean anyPass = false;
				for (int i=0;i<data.size();i++) {
					if (data.get(i).size() > gen) {
						anyPass = true;
						sum += data.get(i).get(gen).bestEver;
						count += 1;
					}
				}
				if (!anyPass) {break outer;}
				System.out.println(gen + "," + (sum/count));
				gen += 1;
			}
	}
	
	/**
	 * Returns a set of columns, one per run. Each column represents the best result
	 * ever achieved by that run so far. With elitism, this is identical to returning
	 * the best in population, but if elitism is set to 0 the results returned may differ.
	 */
	public static void printAllBest(ArrayList<ArrayList<RecordedGeneration>> asGenArrays) {
		int minSize = asGenArrays.get(0).size();
		for (ArrayList<RecordedGeneration> episode:asGenArrays) {
			if (episode.size() < minSize) {
				minSize = episode.size();
			}
		}


		for (int i=0;i<minSize;i++) {
			System.out.print("Gen " + i + ", ");
			for (ArrayList<RecordedGeneration> episode:asGenArrays) {
				int iVal = (int)(1000*episode.get(i).bestEver);
				double dVal = iVal;
				dVal /= 1000;
				System.out.print(" " + dVal);
			}
			System.out.println();
		}
	}
	
	/**
	 * Returns a set of columns, one per run. Each column represents the best result
	 * in the population at that generation.
	 * 
	 */
	public static void printAvrgBest(ArrayList<ArrayList<RecordedGeneration>> asGenArrays) {
		int minSize = asGenArrays.get(0).size();
		for (ArrayList<RecordedGeneration> episode:asGenArrays) {
			if (episode.size() < minSize) {
				minSize = episode.size();
			}
		}


		for (int i=0;i<minSize;i++) {
			double sum = 0;
			double count = 0;
			for (ArrayList<RecordedGeneration> episode:asGenArrays) {
				sum += episode.get(i).bestEver;
				count += 1;
			}
			System.out.println("Gen " + i + ", " + (sum/count));
		}
	}
}
